#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nwIndex.h"

#define CHUNK_ERROR 0
#define CHUNK_DEF 1
#define CHUNK_CALL 2
#define CHUNK_END 3

#define STACK_SIZE 4096
struct chunk *stack[STACK_SIZE] = { 0 };
int stackPtr;

struct checkedLine checkLine(char *line) {
	char *linePtr = line;
	int openTagsFound = 0, closeTagsFound = 0;
	struct checkedLine currentLine = { 0 };

	currentLine.type = CHUNK_ERROR;

	if (*linePtr != '\0' && *linePtr == '@') {
		currentLine.type = CHUNK_END;
		return currentLine;
	}

	while (*linePtr != '\0' && !openTagsFound) {

		if (*linePtr == '<' && *(++linePtr) != '\0' && *linePtr == '<') {
			openTagsFound = 1;
		}

		linePtr++;
	}

	currentLine.chunkNameStart = linePtr;

	while (*linePtr != '\0' && openTagsFound && !closeTagsFound) {

		if (*linePtr == '>' && *(++linePtr) != '\0' && *linePtr == '>') {
			closeTagsFound = 1;
			currentLine.chunkNameEnd = linePtr - 1;
		}
		linePtr++;
	}

	if (*linePtr == '=') {
		currentLine.type = CHUNK_DEF;
	} else if (openTagsFound && closeTagsFound) {
		currentLine.type = CHUNK_CALL;
	}

	return currentLine;
}

void addChild(struct chunk *parent, char *childName) {
	struct chunk *newChild = malloc(sizeof(struct chunk));

	memset(newChild, 0, sizeof(struct chunk));
	strcpy(newChild->name, childName);

	if (parent->child == NULL) {
		parent->child = newChild;
		push(newChild);
	} else {
		struct chunk *childPtr = parent->child;
		while (childPtr->next != NULL) {
			if (strcmp(childPtr->name, childName) == 0) {
				free(newChild);
				return;
			}
			childPtr = childPtr->next;
		}
		childPtr->next = newChild;
		push(newChild);
	}
}

void readChunk(FILE *inputFile, struct chunk *curChunk) {
	char *line = NULL;
	char *startChunk = NULL;
	size_t length = 0;
	ssize_t read;
	struct checkedLine currentLine = { 0 };

	while ((read = getline(&line, &length, inputFile)) != -1) {
		startChunk = line;
		while (startChunk != NULL && (startChunk - line) < length) {
			currentLine = checkLine(startChunk);
			if (currentLine.type == CHUNK_CALL) {
				char *childName = malloc(
						sizeof(char)
								* (currentLine.chunkNameEnd
										- currentLine.chunkNameStart) + 1);
				memset(childName, 0,
						(currentLine.chunkNameEnd - currentLine.chunkNameStart)
								+ 1);
				strncpy(childName, currentLine.chunkNameStart,
						(int) (currentLine.chunkNameEnd
								- currentLine.chunkNameStart));

				addChild(curChunk, childName);

				free(childName);
			} else if (currentLine.type == CHUNK_END) {
				return;
			}
			startChunk = currentLine.chunkNameEnd;
		}

	}
}

void readFileForChunk(FILE *inputFile, struct chunk *curChunk) {

	char *line = NULL;
	size_t length = 0;
	ssize_t read;
	struct checkedLine currentLine = { 0 };

	rewind(inputFile);
	while ((read = getline(&line, &length, inputFile)) != -1) {
		currentLine = checkLine(line);
		if (currentLine.type == CHUNK_DEF) {
			char *childName = malloc(
					sizeof(char)
							* (currentLine.chunkNameEnd
									- currentLine.chunkNameStart) + 1);
			memset(childName, 0,
					(currentLine.chunkNameEnd - currentLine.chunkNameStart)
							+ 1);
			strncpy(childName, currentLine.chunkNameStart,
					(int) (currentLine.chunkNameEnd - currentLine.chunkNameStart));
			if (strcmp(childName, curChunk->name) == 0) {
				readChunk(inputFile, curChunk);
			}
			free(childName);
		}
	}

	free(line);
}

void push(struct chunk *item) {
	stackPtr++;
	if (stackPtr < STACK_SIZE) {
		stack[stackPtr] = item;
	}
}

struct chunk *pop(void) {
	struct chunk *ret = NULL;
	if (stackPtr) {
		ret = stack[stackPtr];
		stackPtr--;
	}
	return ret;
}

void printChunk(struct chunk *curChunk) {
	struct chunk *chunkPtr = curChunk;

	for (int i = 0; i < stackPtr; i++) {
		printf("    ");
	}

	printf("<<%s>>\n", chunkPtr->name);

	if (chunkPtr->child != NULL) {
		push(chunkPtr);
		printChunk(chunkPtr->child);
	} else if (chunkPtr->next != NULL) {
		printChunk(chunkPtr->next);
	} else {
		chunkPtr = pop();
		while (chunkPtr->next == NULL && stackPtr) {
			chunkPtr = pop();
		}
		if (chunkPtr->next != NULL) {
			printChunk(chunkPtr->next);
		}
	}
}

int main(int argc, char **argv) {

	if (argc - 1 < 2 || argc - 1 > 2) {
		printf("Argument Error! Use %s filename startchunk\n", argv[0]);
		return 1;
	}

	char nwFile[256] = { 0 };
	char nwEntry[256] = { 0 };

	strcpy(nwFile, argv[1]);
	strcpy(nwEntry, argv[2]);

	stackPtr = 0;

	FILE *inputFile;
	struct chunk chunkIndex = { 0 };
	struct chunk *curChunk = &chunkIndex;

	strcpy(chunkIndex.name, nwEntry);
	push(curChunk);

	inputFile = fopen(nwFile, "r");

	if (inputFile == NULL) {
		exit(EXIT_FAILURE);
	}

	while ((curChunk = pop()) != NULL) {
		readFileForChunk(inputFile, curChunk);
	}

	printChunk(&chunkIndex);

	fclose(inputFile);

	return 0;
}
